CW_MAX_CHANNELS = 10;
CW_MAX_CHANNEL_ROSTER = 9;

-- Inject a count field into the GuildMicroButton
local CW_GCText = GuildMicroButtonTabard:CreateFontString("GuildMicroButtonTabardCount", "OVERLAY", "NumberFontNormal");
CW_GCText:SetPoint("BOTTOMRIGHT", -4, 4);
CW_GCText:SetJustifyH("RIGHT");
CW_GCText:Hide();

-- Saved Variables
CHATWHODB = {
	showChannels = true,
	showGuild = true
};

-- Localization globals
local L = LibStub("AceLocale-3.0"):GetLocale("ChatWho", true);

local options = {
	name = L["ChatWho"],
	type = 'group',
	args = {
		show_channels = {
			type = 'toggle',
			name = L["Show Channels"],
			desc = L["SHOWCHANNELS_DESC"],
			descStyle = "inline",
			width = "full",
			set = function(info, enabled)
					CHATWHODB["showChannels"] = enabled;
					if (enabled) then
						ChatWho_ShowChannels();
					else
						ChatWho_HideChannels();
					end
					if (info[0]) then
						LibStub("AceConsole-3.0"):Print(enabled and L["SHOWCHANNELS_ENABLED"] or L["SHOWCHANNELS_DISABLED"]);
					end
				 end,
			get = function(info) return CHATWHODB["showChannels"] end,
		},
		show_guild = {
			type = 'toggle',
			name = L["Show Guild"],
			desc = L["SHOWGUILD_DESC"],
			descStyle = "inline",
			width = "full",
			set = function(info, enabled)
					CHATWHODB["showGuild"] = enabled;
					if ( enabled and IsInGuild() ) then
						CW_GCText:Show();
					else
						CW_GCText:Hide();
					end
					if (info[0]) then
						LibStub("AceConsole-3.0"):Print(enabled and L["SHOWGUILD_ENABLED"] or L["SHOWGUILD_DISABLED"]);
					end
				 end,
			get = function(info) return CHATWHODB["showGuild"] end,
		},
	},
};
LibStub("AceConfig-3.0"):RegisterOptionsTable("ChatWho", options, {"chatwhox"});
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ChatWho", L["ChatWho"]);

-- This comes from Choonster @ Frostmourne: http://us.battle.net/wow/en/forum/topic/4044466747
local CW_onlineMatchString = ERR_FRIEND_ONLINE_SS:format(".+", "(.+)%"):gsub("%[","%%[");
local CW_offlineMatchString = ERR_FRIEND_OFFLINE_S:format("(.+)");
local CW_joinGuildMatchString = ERR_GUILD_JOIN_S:format("(.+)");
local CW_leftGuildMatchString = ERR_GUILD_LEAVE_S:format("(.+)");

--[[
	Guild Support functions
--]]
local CW_GuildRoster = {};
local CW_OnlineMembers = 0;
function ChatWho_GuildRosterUpdate()
	wipe(CW_GuildRoster);
	if IsInGuild() then
		local totalMembers, onlineMembers, onlineAndMobileMembers = GetNumGuildMembers();
		CW_GCText:SetText(onlineMembers);
	
		local index = 1;
		local insert = table.insert;
		
		for i = 1, onlineAndMobileMembers do
			local name, rank, rankIndex, level, class, zone, note, officernote, online, isAway, classFileName, achievementPoints, achievementRank, isMobile, canSoR, repStanding = GetGuildRosterInfo(i);
			if (online) then
				insert(CW_GuildRoster, name);
				index = index + 1;
				if (index > min(onlineMembers, CW_MAX_CHANNEL_ROSTER)) then
					break;
				end
			end
		end
		CW_OnlineMembers = onlineMembers;
		if ( CHATWHODB["showGuild"] ) then
			CW_GCText:Show();
		end
	else
		CW_OnlineMembers = 0;
		CW_GCText:Hide();
	end
end

function ChatWho_GuildMicroButtonOnEnter()
	if ( CHATWHODB["showGuild"] and IsInGuild() ) then
		GameTooltip:AddLine( table.concat(CW_GuildRoster, "\n") );
		if (CW_OnlineMembers > CW_MAX_CHANNEL_ROSTER) then
			GameTooltip:AddLine("...");
		end
		GameTooltip:Show();
	end
end

--[[
	Channel Support functions
--]]
local CW_Channels = {}; -- Array of channel structures allowing much reuse.
local CW_WorkQueue = {};
local CW_Updating = false;
local CW_Paused = false;
function ChatWho_QueueChannelInfo( channelID )
	if ( not CW_Channels[channelID].used ) then  return;  end

	-- Ignore entries already queued
	for i, v in pairs(CW_WorkQueue) do
		if (v == channelID) then
			return;
		end;
	end

	if( not CW_Updating and not CW_Paused) then
		ChatWho_ChannelQueryServer( channelID );
	else
		table.insert(CW_WorkQueue, channelID);
	end
	-- Regardless of the previous check. We should always be waiting for an update after this
	CW_Updating = true;
end

function ChatWho_HandleChannelInfo( displayID )
	-- We may get server responses we didn't request.
	-- We are A-OK with this. The info is still good, and possibly more up to date.
	local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo( displayID );

	if ( category and category == "CHANNEL_CATEGORY_CUSTOM") then
		local channel = CW_Channels[channelNumber];
		channel.name = name;
		channel.displayID = displayID;
		channel.count = count;
		channel.used = true;
		
		local roster = channel.roster;
		for i=1, min(count, CW_MAX_CHANNEL_ROSTER) do
			roster[i] = GetChannelRosterInfo(displayID, i);
		end
	else
		CW_Channels[channelNumber].used = false;
	end
	
	ChatWho_UpdateButtons();
	ChatWho_ChannelQueryNext();
end

function ChatWho_ChannelQueryNext()
	if ( CW_Updating and not CW_Paused) then
		local nextChannelID = table.remove(CW_WorkQueue, 1);
		if ( nextChannelID ) then
			ChatWho_ChannelQueryServer( nextChannelID );
		else
			CW_Updating = false;
		end
	end
end

local CW_LastID = 0;
function ChatWho_ChannelQueryServer( queryID )
	if ( not CW_Channels[queryID].used ) then
		-- Just in case our Queue was off
		-- or a channel got queued and deleted while we were paused.
		ChatWho_ChannelQueryNext();
	end
	
	-- If the last query was asked recently. Toggle the question quickly to ask the server again
	local queryDisplayID = CW_Channels[queryID].displayID;
	
	if ( CW_LastID < 1 or queryDisplayID == CW_LastID ) then
		for i=1, GetNumDisplayChannels() do
			local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(i);
			if ( channelNumber and i ~= queryDisplayID) then
				SetSelectedDisplayChannel( i );
				break;
			end
		end
	end
	
	SetSelectedDisplayChannel( queryDisplayID );
	CW_LastID = queryDisplayID;
end

function ChatWho_ChannelPause()
	CW_Paused = true;
end

function ChatWho_ChannelUnpause()
	if ( not CW_Paused ) then return; end
	CW_Paused = false;
	ChatWho_ChannelQueryNext();
end

function ChatWho_SetButtonSide(chatFrame, buttonSide)
	ChatWhoButton1:ClearAllPoints();
	if (buttonSide == "right") then		
		ChatWhoButton1:SetPoint("TOPLEFT", FriendsMicroButton, "TOPRIGHT", -2, 0);
	else
		ChatWhoButton1:SetPoint("TOPRIGHT", FriendsMicroButton, "TOPLEFT", 2, 0);
	end
end

local origLeft, origRight, origTop, origBottom, origXOFfset;
function ChatWho_ShowChannels()
	-- The following controls the fixed offset of the chat frame: (32 is the DEFAULT, from UIParent.lua)
	-- And default clamping is 	:SetClampRectInsets(-35, 35, 26, -50);
	UIPARENT_MANAGED_FRAME_POSITIONS[DEFAULT_CHAT_FRAME:GetName()].xOffset = 60;
	DEFAULT_CHAT_FRAME:SetClampRectInsets(-62, 62, 26, -50);

	UIParent_ManageFramePositions();
	ChatWhoFrame:Show();
end

function ChatWho_HideChannels()
	UIPARENT_MANAGED_FRAME_POSITIONS[DEFAULT_CHAT_FRAME:GetName()].xOffset = origXOFfset;
	DEFAULT_CHAT_FRAME:SetClampRectInsets(origLeft, origRight, origTop, origBottom);
	
	ChatWhoFrame:Hide();
	UIParent_ManageFramePositions();
end

function ChatWho_OnAddonLoaded(self, event, arg1, ...)
	if (arg1 ~= "ChatWho") then return; end
	self:UnregisterEvent("ADDON_LOADED");
	
	origLeft, origRight, origTop, origBottom = DEFAULT_CHAT_FRAME:GetClampRectInsets();
	origXOffset = UIPARENT_MANAGED_FRAME_POSITIONS[DEFAULT_CHAT_FRAME:GetName()].xOffset;
	
	if ( CHATWHODB["showChannels"] ) then
		ChatWho_ShowChannels();
	else
		ChatWho_HideChannels();
	end

	if ( CHATWHODB["showGuild"] ) then
		CW_GCText:Show();
	else
		CW_GCText:Hide();
	end
end

-- CHAT_MSG_CHANNEL_NOTICE - For player leaving/joining a channel. Superseded by CHANNEL_UI_UPDATE.
function ChatWho_OnLoad(self, ...)
	self:RegisterEvent("GUILD_ROSTER_UPDATE");
	self:RegisterEvent("PLAYER_GUILD_UPDATE");
	self:RegisterEvent("CHANNEL_ROSTER_UPDATE");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("CHAT_MSG_CHANNEL_JOIN");
	self:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE");
	self:RegisterEvent("CHAT_MSG_SYSTEM");
	self:RegisterEvent("CHANNEL_UI_UPDATE");
	self:RegisterEvent("ADDON_LOADED");
	
	hooksecurefunc("FCF_SetButtonSide", ChatWho_SetButtonSide);

	ChannelFrame:HookScript("OnShow", ChatWho_ChannelPause);
	ChannelFrame:HookScript("OnHide", ChatWho_ChannelUnpause);
	GuildMicroButton:HookScript("OnEnter", ChatWho_GuildMicroButtonOnEnter);

	for i=2, CW_MAX_CHANNELS do
		local button = CreateFrame("Button", "ChatWhoButton"..i, ChatWhoFrame, "ChatWhoButtonTemplate");
		button:SetPoint("TOPRIGHT", _G["ChatWhoButton"..(i-1)], "BOTTOMRIGHT", 0, -4);
		button:GetNormalTexture():SetDesaturated(true);
	end
	ChatWhoButton1:GetNormalTexture():SetDesaturated(true);
	
	-- Channels are numbered as per warcraft client
	for i=1, CW_MAX_CHANNELS do
		CW_Channels[i] = {
			number = i,
			name = nil,
			displayID = nil, -- ID used by the ChannelFrame UI, this is channel number + some value 3 or higher
			used = false,
			count = 0,
			roster = { UnitName("Player") },
			tag = "CHANNEL"..i
		};
	end
end

function ChatWho_OnEvent(self, event, ...)
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13, arg14 = ...;

	if ( event == "GUILD_ROSTER_UPDATE" or event == "PLAYER_GUILD_UPDATE" ) then
		ChatWho_GuildRosterUpdate();

	elseif ( event == "CHAT_MSG_SYSTEM" ) then
		-- Other things may trigger the GuildRoster() call, but just in case..
		if ( IsInGuild() and
			(arg1:find(CW_onlineMatchString) or arg1:find(CW_offlineMatchString) or
			 arg1:find(CW_joinGuildMatchString) or arg1:find(CW_leftGuildMatchString))
			) then
			GuildRoster();
		end

	elseif ( event == "CHANNEL_ROSTER_UPDATE" ) then
		ChatWho_HandleChannelInfo( arg1 );

	elseif ( strsub(event, 1, 8) == "CHAT_MSG" ) then
		-- Other players join/leave channel
		if (CW_Channels[arg8].used) then
			ChatWho_QueueChannelInfo( arg8 );
			if (event == "CHAT_MSG_CHANNEL_JOIN" and CW_Channels[arg8].used) then
				CW_Channels[arg8].button.playerJoined:Play();
			end
		end

	elseif ( event == "PLAYER_ENTERING_WORLD" or event == "CHANNEL_UI_UPDATE" ) then
		-- Occurs before channels are joined,
		-- but we don't rejoin channels when we /reload
		-- So this helps

		-- Try not to overdue calls to this.
		-- For instance Login is 1 PLAYER_ENTERING_WORLD and 10+ CHANNEL_UI_UPDATE 
		if (not CW_Updating or CW_Paused) then
			ChatWho_UpdateAllChannels( );
		end

	elseif ( event == "ADDON_LOADED" ) then
		ChatWho_OnAddonLoaded(self, event, ...);
	end
end

local firstTime = true;
function ChatWho_UpdateAllChannels()
	-- We might not be told of removed channels. Sweep and correct
	for i=1, CW_MAX_CHANNELS do
		CW_Channels[i].used = false;
	end

	-- If the last Display entry is a header, force it to be open.
	local updateButtons = true;
	if( select(2, GetChannelDisplayInfo( GetNumDisplayChannels() )) ) then
		ExpandChannelHeader( GetNumDisplayChannels() );
		updateButtons = false; -- Skip updating, the subsequent call for a UI Update will get them
	end
	
	-- ChannelFrame sometimes gets confused, prevent it from showing its DropDown
	ChannelListDropDown.clicked = nil;

	local withQuery = firstTime;
	for i=1, GetNumDisplayChannels() do
		local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(i);
		if ( category and category == "CHANNEL_CATEGORY_CUSTOM") then
			local channel = CW_Channels[channelNumber];
			channel.name = name;
			channel.displayID = i;
			if ( count ) then
				channel.count = count;
			else
				withQuery = true;
			end
			channel.used = true;

			if ( withQuery ) then
				ChatWho_QueueChannelInfo( channelNumber );
			end
		end
	end
	
	if (updateButtons) then
		ChatWho_UpdateButtons();
	end
	firstTime = false;
end

function ChatWho_UpdateButtons()
	local i = 1;

	for k=1, CW_MAX_CHANNELS do
		local channel = CW_Channels[k];
		if ( channel.used ) then
			local info = ChatTypeInfo[channel.tag];
			local button = _G["ChatWhoButton"..i];
			button.channel = channel;
			button.tint:SetTexture(info.r, info.g, info.b);
			if (channel.count) then
				_G["ChatWhoButton"..i.."Count"]:SetText(channel.count);
			end
			button:Show();
			channel.button = button;
			i = i + 1
		end
	end
	
	for k = i, CW_MAX_CHANNELS do
		_G["ChatWhoButton"..k]:Hide();
	end
end
